
' Model credits at end of source.

#Import "<std>"
#Import "<mojo>"
#Import "<mojo3d>"
#Import "<mojo3d-loaders>"

#Import "aerialcamera"
#Import "assets/"

Using std..
Using mojo..
Using mojo3d..

Class IslandDemo Extends Window
	
	' Set this to False if it runs slowly on low-end devices!
	
	Const GROUND_SHADOWS:Bool		= True
	
	Const WINDOW_WIDTH:Int			= 640
	Const WINDOW_HEIGHT:Int			= 480
	Const WINDOW_FLAGS:WindowFlags	= WindowFlags.Resizable

'	Const WINDOW_WIDTH:Int			= 1920
'	Const WINDOW_HEIGHT:Int			= 1080
'	Const WINDOW_FLAGS:WindowFlags	= WindowFlags.Fullscreen

	Field scene:Scene
	Field camera:AerialCamera
	Field plane:PlaneBehaviour
	
	Method New (title:String = "Island Demo", width:Int = WINDOW_WIDTH, height:Int = WINDOW_HEIGHT, flags:WindowFlags = WINDOW_FLAGS)
		Super.New (title, width, height, flags)
	End
	
	Method OnCreateWindow () Override

		' ---------------------------------------------------------------------
		' Scene setup:
		' ---------------------------------------------------------------------
		' Gets a reference to the mojo3d default scene and sets a few options...
		' ---------------------------------------------------------------------
		
		scene									= Scene.GetCurrent ()

			scene.ClearColor					= New Color (0.2, 0.6, 1.0)
			scene.AmbientLight					= scene.ClearColor * 0.25
			scene.FogColor						= scene.ClearColor
			scene.FogNear						= 128
			scene.FogFar						= 2048

		' ---------------------------------------------------------------------
		' Camera setup:
		' ---------------------------------------------------------------------
		' Creates an AerialCamera from the imported class defined in aerialcamera.monkey2...
		' ---------------------------------------------------------------------
		
		camera									= New AerialCamera (App.ActiveWindow.Rect, Null, 4096)
		
		' ---------------------------------------------------------------------
		' Light setup:
		' ---------------------------------------------------------------------
		' Creates a light representing the Sun. The default light type is directional,
		' a universal 'wall' of light pointing in a given direction...
		' ---------------------------------------------------------------------
		
		Local light:Light						= New Light

			light.CastsShadow					= True
	
			light.Rotate (45, 45, 0)
		
		' ---------------------------------------------------------------------
		' Ground model and physics setup:
		' ---------------------------------------------------------------------
		' Loads an external model and adds a physics collider and rigid body...
		' ---------------------------------------------------------------------
		
		Local ground_size:Float					= 4096 * 0.5
		Local ground_box:Boxf					= New Boxf (-ground_size, -ground_size * 0.5, -ground_size, ground_size, 0, ground_size)
		Local ground_model:Model				= Model.Load ("asset::model_gltf_6G3x4Sgg6iX_7QCCWe9sgpb\model.gltf")'CreateBox( groundBox,1,1,1,groundMaterial )

			ground_model.CastsShadow			= GROUND_SHADOWS
			
			ground_model.Mesh.FitVertices (ground_box, False)
			
			For Local mat:Material = Eachin ground_model.Materials
				mat.CullMode = CullMode.Back	
			Next

		Local ground_collider:MeshCollider		= ground_model.AddComponent <MeshCollider> ()

			ground_collider.Mesh				= ground_model.Mesh
		
		Local ground_body:RigidBody				= ground_model.AddComponent <RigidBody> ()

			ground_body.Mass					= 0
	
		' ---------------------------------------------------------------------
		' Plane setup...
		' ---------------------------------------------------------------------
		' Loads a plane model and set up physics behaviour...
		' ---------------------------------------------------------------------

		' Model setup...

 		' Plane's dimensions and mass are based on SEPECAT Jaguar, per Wikipedia data:
 		
 		' https://en.wikipedia.org/wiki/SEPECAT_Jaguar
 		
 		' Wingspan:		16.83m
 		' Loaded mass:	10954kg
 		
		Local plane_size:Float					= 16.83 * 0.5
		Local plane_box:Boxf					= New Boxf (-plane_size, -plane_size, -plane_size, plane_size, plane_size, plane_size)
		
		Local plane_model:Model					= Model.Load ("asset::1397 Jet_gltf_3B3Pa6BHXn1_fKZwaiJPXpf\1397 Jet.gltf")

			plane_model.Mesh.FitVertices (plane_box)
			plane_model.Mesh.Rotate (0, 180, 0)
			
			plane_model.Move (0.0, 5.0, 0.0)

		' Cull hidden tris...
		
		For Local mat:Material = Eachin plane_model.Materials
			mat.CullMode = CullMode.Back	
		Next

		' Physics setup...
		
		Local plane_collider:SphereCollider		= plane_model.AddComponent <SphereCollider> ()

			plane_collider.Radius				= plane_size * 0.6
		
		Local plane_body:RigidBody				= plane_model.AddComponent <RigidBody> ()

			plane_body.Mass						= 10954.0
			plane_body.Restitution				= 0.5
			plane_body.AngularDamping			= 0.9
			plane_body.LinearDamping			= 0.5
			plane_body.Friction					= 0.0
		
		plane									= New PlaneBehaviour (plane_model)

		' Debug sphere for visibility of collision radius only...
		
		Local coll_vis:Bool = False
		
		If coll_vis

			' Create a sphere model used to represent the collision sphere; hidden by default...
	
			Local plane_collider_vis:Model					= Model.CreateSphere (plane_collider.Radius, 16, 16, New PbrMaterial (Color.White), plane_model)
	
				plane_collider_vis.Alpha					= 0.25
		
		End
		
		' Hide mouse pointer...
		
		Mouse.PointerVisible					= False
		
	End
	
	Method OnRender (canvas:Canvas) Override
	
		If Keyboard.KeyHit (Key.Escape) Then App.Terminate ()
		
		RequestRender ()
		scene.Update ()
		
		camera.Update (plane)
		camera.Render (canvas)
		
		canvas.DrawText ("FPS: " + App.FPS, 0, 0)
		canvas.DrawText ("Y: " + plane.Entity.Y, 0, 20)
		
	End
	
End

' Plane's physics...

Class PlaneBehaviour Extends Behaviour
	
	Property RollRate:Float ()
		Return roll_rate
	End
	
	Property PitchRate:Float ()
		Return pitch_rate
	End
	
	Field plane_body:RigidBody

	Field roll_rate:Float	= 500000.0
	Field pitch_rate:Float	= 200000.0

	Field throttle:Float	= 1000000.0
	
	'	Throttle: 22750.0 * 2.0 Newtons would be correct figure for max dry thrust, so
	'	not sure why 1 million is needed here! Possibly affected by LinearDamping value?
	
	'	From wiki link above, under Powerplant:
	
	'	Dry thrust: 22.75 kN (5,115 lbf) each
	'	Thrust with afterburner: 32.5 kN (7,305 lbf) each
	
	Method New (entity:Entity)
		
		Super.New (entity)
		AddInstance ()

	End
	
	Method OnStart () Override

		' Storing this just for typing convenience...
		
		plane_body = Entity.GetComponent <RigidBody> ()

		' Give it a shove (instantly applied, which is
		' not physically possible), to get it started...
		
		plane_body.ApplyImpulse (Entity.Basis * New Vec3f (0.0, 0.0, 500000.0))

	End
	
	Method OnUpdate (elapsed:Float) Override

		' -Scene.GetCurrent ().World.Gravity.Y * 1.0

		plane_body.ApplyForce (Entity.Basis * New Vec3f (0.0, 0.0, throttle))
		
		If Keyboard.KeyDown (Key.A)
			throttle = throttle + 10000.0
		Endif
		
		If Keyboard.KeyDown (Key.Z)
			throttle = throttle - 10000.0
		Endif
		
		If Keyboard.KeyDown (Key.Left)
			plane_body.ApplyTorque (Entity.Basis * New Vec3f (0.0, 0.0, RollRate))
		Endif

		If Keyboard.KeyDown (Key.Right)
			plane_body.ApplyTorque (Entity.Basis * New Vec3f (0.0, 0.0, -RollRate))
		Endif

		If Keyboard.KeyDown (Key.Up)
			plane_body.ApplyTorque (Entity.Basis * New Vec3f (PitchRate, 0.0, 0.0))
		Endif

		If Keyboard.KeyDown (Key.Down)
			plane_body.ApplyTorque (Entity.Basis * New Vec3f (-PitchRate, 0.0, 0.0))
		Endif
		
	End
	
End

' Support functions...

Const RAD_DIVIDER:Float = 180.0 / Pi
Const DEG_DIVIDER:Float = Pi / 180.0

Function Degrees:Float (radians:Float)
	Return radians * RAD_DIVIDER
End

Function Radians:Float (degrees:Float)
	Return degrees * DEG_DIVIDER
End

Function TransformRange:Float (input_value:Float, from_min:Float, from_max:Float, to_min:Float, to_max:Float)

	' Algorithm via jerryjvl at https://stackoverflow.com/questions/929103/convert-a-number-range-to-another-range-maintaining-ratio
	
	Local from_delta:Float	= from_max	- from_min	' Input range,	eg. 0.0 - 1.0
	Local to_delta:Float	= to_max	- to_min	' Output range,	eg. 5.0 - 10.0
	
	Assert (from_delta <> 0.0, "TransformRange: Invalid input range!")
	
	Return (((input_value - from_min) * to_delta) / from_delta) + to_min
	
End

Function Blend:Float (in:Float, target:Float, delta:Float = 0.1)
	If Abs (target - in) < Abs (delta) Then Return target
	Return in + ((target - in) * delta)
End

' Needed to rotate loaded mesh...

Class Mesh Extension

	' Thanks, DoctorWhoof!

	Method Rotate (pitch:Float, yaw:Float, roll:Float)
		TransformVertices (AffineMat4f.Rotation (Radians (pitch), Radians(yaw), Radians (roll)))
	End
	
End

Function Main ()

	New AppInstance
	New IslandDemo
	
	App.Run ()

End

' Model credits -- thanks to the authors!

' Unmodified GLTF models from sources below.

' Model: "A little Irish island"

' 		Author

'			John Huikku:	https://poly.google.com/user/1b9F61Bj5l9
'			Source:			https://poly.google.com/view/6G3x4Sgg6iX

'		License

'			Attribution 3.0 Unported (CC BY 3.0) aka. "CC-BY 3.0"
'			https://creativecommons.org/licenses/by/3.0/
'			https://support.google.com/poly/answer/7418679

' Model: "Jet"

'		Author

'			Poly by Google:	https://poly.google.com/user/1b9F61Bj5l9
'			Source:			https://poly.google.com/view/3B3Pa6BHXn1

'		License

'			Attribution 3.0 Unported (CC BY 3.0) aka. "CC-BY 3.0"
'			https://creativecommons.org/licenses/by/3.0/
'			https://support.google.com/poly/answer/7418679
